//
//  JVFloatLabeledTextField.m
//  JVFloatLabeledTextField
//
//  The MIT License (MIT)
//
//  Copyright (c) 2013 Jared Verdi
//  Original Concept by Matt D. Smith
//  http://dribbble.com/shots/1254439--GIF-Mobile-Form-Interaction?list=users
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy of
//  this software and associated documentation files (the "Software"), to deal in
//  the Software without restriction, including without limitation the rights to
//  use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
//  the Software, and to permit persons to whom the Software is furnished to do so,
//  subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in all
//  copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
//  FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
//  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
//  IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
//  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#import "JVFloatLabeledTextField.h"
#import "NSString+TextDirectionality.h"
#import "Constants.h"

#define kFloatingLabelShowAnimationDuration 0.3f
#define kFloatingLabelHideAnimationDuration 0.3f

@implementation JVFloatLabeledTextField {
    NSString *placeHolderText;
    NSAttributedString *attributedPlaceholderStr;
    
    BOOL shouldReturn;
}

- (instancetype)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        [self commonInit];
    }
    shouldReturn = NO;
    return self;
}

- (instancetype)initWithCoder:(NSCoder *)aDecoder
{
    self = [super initWithCoder:aDecoder];
    if (self) {
        [self commonInit];
    }
    return self;
}

- (void)commonInit
{
    _floatingLabel = [LocalisedLabel new];
    _floatingLabel.alpha = 0.0f;
    [self addSubview:_floatingLabel];
    // some basic default fonts/colors
    _floatingLabelFont = [UIFont fontWithName:@"SFUIText-Regular" size:11.0f];
    _floatingLabel.font = _floatingLabelFont;
    _floatingLabelActiveTextColor = [self labelActiveColor];
    self.tintColor = [self labelActiveColor];
    _animateEvenIfNotFirstResponder = NO;
    _floatingLabelShowAnimationDuration = kFloatingLabelShowAnimationDuration;
    _floatingLabelHideAnimationDuration = kFloatingLabelHideAnimationDuration;
    [self setFloatingLabelText:[NSString stringWithFormat:@"%@", self.placeholder]];
}

#pragma mark -

- (UIColor *)labelActiveColor
{
    return barney_purple;
}

- (void)setFloatingLabelFont:(UIFont *)floatingLabelFont
{
    _floatingLabelFont = floatingLabelFont;
    _floatingLabel.font = (_floatingLabelFont ? _floatingLabelFont : [UIFont fontWithName:@"Roboto-Regular" size:11.0]);
    [self setFloatingLabelText:self.placeholder];
}

- (void)showFloatingLabel:(BOOL)animated
{
    void (^showBlock)() = ^{
        _floatingLabel.alpha = 1.0f;
        _floatingLabel.frame = CGRectMake(_floatingLabel.frame.origin.x,
                                          0,
                                          _floatingLabel.frame.size.width,
                                          _floatingLabel.frame.size.height);
    };
    
    if (animated || _animateEvenIfNotFirstResponder) {
        [UIView animateWithDuration:_floatingLabelShowAnimationDuration
                              delay:0.0f
                            options:UIViewAnimationOptionBeginFromCurrentState | UIViewAnimationOptionCurveEaseOut
                         animations:showBlock
                         completion:nil];
    }
    else {
        showBlock();
    }
}

- (void)hideFloatingLabel:(BOOL)animated
{
    void (^hideBlock)() = ^{
        _floatingLabel.alpha = 0.0f;
        _floatingLabel.frame = CGRectMake(_floatingLabel.frame.origin.x,
                                          _floatingLabel.font.lineHeight + _placeholderYPadding +2,
                                          _floatingLabel.frame.size.width,
                                          _floatingLabel.frame.size.height);
        
    };
    
    if (animated || _animateEvenIfNotFirstResponder) {
        [UIView animateWithDuration:_floatingLabelHideAnimationDuration
                              delay:0.0f
                            options:UIViewAnimationOptionBeginFromCurrentState | UIViewAnimationOptionCurveEaseIn
                         animations:hideBlock
                         completion:nil];
    }
    else {
        hideBlock();
    }
}

- (void)setLabelOriginForTextAlignment
{
    CGRect textRect = [self textRectForBounds:self.bounds];
    
    CGFloat originX = textRect.origin.x;
    
    if (self.textAlignment == NSTextAlignmentCenter) {
        originX = textRect.origin.x + (textRect.size.width/2) - (_floatingLabel.frame.size.width/2);
    }
    else if (self.textAlignment == NSTextAlignmentRight) {
        originX = textRect.origin.x + textRect.size.width - _floatingLabel.frame.size.width;
    }
    else if (self.textAlignment == NSTextAlignmentNatural) {
        JVTextDirection baseDirection = [_floatingLabel.text getBaseDirection];
        if (baseDirection == JVTextDirectionRightToLeft) {
            originX = textRect.origin.x + textRect.size.width - _floatingLabel.frame.size.width;
        }
    }
    
    _floatingLabel.frame = CGRectMake(originX, _floatingLabel.frame.origin.y,
                                      _floatingLabel.frame.size.width, _floatingLabel.frame.size.height);
}

- (void)setFloatingLabelText:(NSString *)text
{
    _floatingLabel.text = text;
    [self setNeedsLayout];
}

#pragma mark - UITextField

- (void)setPlaceholder:(NSString *)placeholder
{
    [super setPlaceholder:placeholder];
    //    placeHolderText = [NSString stringWithString:placeholder];
    if(attributedPlaceholderStr.string != nil)
        [self setFloatingLabelText:placeHolderText];
    else
        [self setFloatingLabelText:placeholder];
}

- (void)setAttributedPlaceholder:(NSAttributedString *)attributedPlaceholder
{
    [super setAttributedPlaceholder:attributedPlaceholder];
    if(attributedPlaceholderStr.string != nil)
        [self setFloatingLabelText:attributedPlaceholderStr.string];
    else
        [self setFloatingLabelText:attributedPlaceholder.string];
}

- (void)setPlaceholder:(NSString *)placeholder floatingTitle:(NSString *)floatingTitle
{
    [super setPlaceholder:placeholder];
    [self setFloatingLabelText:floatingTitle];
}

- (CGRect)textRectForBounds:(CGRect)bounds
{
    CGRect rect = [super textRectForBounds:bounds];
    //    if ([self.text length]) {
    CGFloat topInset = ceilf(_floatingLabel.font.lineHeight + _placeholderYPadding);
   topInset = MIN(topInset, [self maxTopInset]);

    rect = UIEdgeInsetsInsetRect(rect, UIEdgeInsetsMake(topInset, 0.0f, 0.0f, 0.0f));
    //    }
    return CGRectIntegral(rect);
}

- (CGRect)editingRectForBounds:(CGRect)bounds
{
    CGRect rect = [super editingRectForBounds:bounds];
    //    if ([self.text length]) {
    CGFloat topInset = ceilf(_floatingLabel.font.lineHeight + _placeholderYPadding);
    topInset = MIN(topInset, [self maxTopInset]);

    rect = UIEdgeInsetsInsetRect(rect, UIEdgeInsetsMake(topInset, 0.0f, 0.0f, 0.0f));
    //    }
    return CGRectIntegral(rect);
}

- (CGRect)clearButtonRectForBounds:(CGRect)bounds
{
    CGRect rect = [super clearButtonRectForBounds:bounds];
    if ([self.text length]) {
        CGFloat topInset = ceilf(_floatingLabel.font.lineHeight + _placeholderYPadding);
        topInset = MIN(topInset, [self maxTopInset]);
        rect = CGRectMake(rect.origin.x, rect.origin.y + topInset / 2.0f, rect.size.width, rect.size.height);
    }
    return CGRectIntegral(rect);
}

- (CGFloat)maxTopInset
{
    return MAX(0, floorf(self.bounds.size.height - self.font.lineHeight ));
}

- (void)setTextAlignment:(NSTextAlignment)textAlignment
{
    [super setTextAlignment:textAlignment];
    [self setNeedsLayout];
}

- (void)layoutSubviews
{
    [super layoutSubviews];
//    if(shouldReturn) {
//        shouldReturn = NO;
//        return;
//    }
    
    [self setLabelOriginForTextAlignment];
    
    if (self.floatingLabelFont) {
        _floatingLabel.font = self.floatingLabelFont;
    }
    BOOL firstResponder = self.isFirstResponder;
    //if background is dark, floating label color will be white with 50% opacity
        if(self.isBgDark)
            _floatingLabelTextColor = [UIColor colorWithRed:255.0/255.0 green:255.0/255.0 blue:255.0/255.0 alpha:0.50];
    //if background is light, floating label color will be charcoalGrey with 100% opacity
        else
            _floatingLabelTextColor = [UIColor colorWithRed:48/255.0 green:54/255.0 blue:57/255.0 alpha:1.0];
    [_floatingLabel sizeToFit];
    
    _floatingLabel.textColor = (firstResponder && self.text ?
                                    self.floatingLabelActiveTextColor : self.floatingLabelTextColor);
    
    
    
    if (!firstResponder && (![self.placeholder isEqualToString:@""] || ![self.attributedPlaceholder.string isEqualToString:@""])) {
        //shouldReturn = YES;
        [self hideFloatingLabel:firstResponder];
    }
    else if (firstResponder && (![self.placeholder isEqualToString:@""] || ![self.attributedPlaceholder.string isEqualToString:@""])){
            placeHolderText = [NSString stringWithString:self.placeholder];
            
            NSDictionary *attributesFromString = [self.attributedPlaceholder attributesAtIndex:0 longestEffectiveRange:nil inRange:NSMakeRange(0, self.attributedPlaceholder.length)];
            
            attributedPlaceholderStr = [[NSAttributedString alloc]initWithString:self.attributedPlaceholder.string attributes: attributesFromString];
            self.placeholder = @"";
            self.attributedPlaceholder = [[NSAttributedString alloc]initWithString:@"" attributes: @{}];
            //shouldReturn = YES;
            [self showFloatingLabel:firstResponder];
        }
    if (!firstResponder && [self.text length]) {
        [self showFloatingLabel:firstResponder];
    }
    else if (!firstResponder && ([self.placeholder isEqualToString:@""] || [self.attributedPlaceholder.string isEqualToString:@""])) {
        self.placeholder = placeHolderText;
        self.attributedPlaceholder = attributedPlaceholderStr;
        if([self.text length]) {
            [self showFloatingLabel:YES];
        }
        else {
            [self hideFloatingLabel:firstResponder];
        }
        //shouldReturn = YES;
    }
}

@end

